/*
 * ToRegExpStateElimination.cpp
 *
 *  Created on: 9. 2. 2014
 *	  Author: Tomas Pecka
 */

#include "ToRegExpStateElimination.h"

#include "../../regexp/simplify/RegExpOptimize.h"
#include "../../regexp/transform/RegExpAlternate.h"
#include "../../regexp/transform/RegExpConcatenate.h"
#include "../../regexp/transform/RegExpIterate.h"

#include <common/createUnique.hpp>

#include <label/FinalStateLabel.h>

namespace automaton {

namespace convert {

regexp::RegExp ToRegExpStateElimination::convert(const automaton::Automaton& automaton) {
	return dispatch(automaton.getData());
}

template<class T>
regexp::UnboundedRegExp < > ToRegExpStateElimination::convert(const T& automaton) {
	if(automaton.getFinalStates().size() == 0)
		return regexp::UnboundedRegExp < > (regexp::UnboundedRegExpStructure < alphabet::Symbol > (regexp::UnboundedRegExpEmpty < alphabet::Symbol > ()));

	// steps 1 + 2
	automaton::ExtendedNFA < > extendedAutomaton(automaton);
	extendExtendedNFA(extendedAutomaton);

	// step 3 - Exterminate!
	// select all states that are neither final nor initial
	std::set<label::Label> statesToEliminate = extendedAutomaton.getStates();
	statesToEliminate.erase(extendedAutomaton.getInitialState());
	statesToEliminate.erase(*extendedAutomaton.getFinalStates().begin());

	for(const auto& state : statesToEliminate)
		extendedAutomaton = eliminateState(extendedAutomaton, state);

	// step 4
	regexp::UnboundedRegExpStructure < alphabet::Symbol > finalStateLoop = regexp::RegExpIterate::iterate ( transition(extendedAutomaton, *extendedAutomaton.getFinalStates().begin(), *extendedAutomaton.getFinalStates().begin()));
	regexp::UnboundedRegExpStructure < alphabet::Symbol > initialToFinalState = regexp::RegExpConcatenate::concatenate ( transition(extendedAutomaton, extendedAutomaton.getInitialState(), *extendedAutomaton.getFinalStates().begin()),finalStateLoop );
	return regexp::UnboundedRegExp < > (regexp::simplify::RegExpOptimize::optimize(initialToFinalState));
}

auto ToRegExpStateEliminationEpsilonNFA = ToRegExpStateElimination::RegistratorWrapper<regexp::UnboundedRegExp < >, automaton::EpsilonNFA < > >(ToRegExpStateElimination::convert);
auto ToRegExpStateEliminationMultiInitialStateNFA = ToRegExpStateElimination::RegistratorWrapper<regexp::UnboundedRegExp < >, automaton::MultiInitialStateNFA < > >(ToRegExpStateElimination::convert);
auto ToRegExpStateEliminationNFA = ToRegExpStateElimination::RegistratorWrapper<regexp::UnboundedRegExp < >, automaton::NFA < > >(ToRegExpStateElimination::convert);
auto ToRegExpStateEliminationDFA = ToRegExpStateElimination::RegistratorWrapper<regexp::UnboundedRegExp < >, automaton::DFA<>>(ToRegExpStateElimination::convert);
auto ToRegExpStateEliminationExtendedNFA = ToRegExpStateElimination::RegistratorWrapper<regexp::UnboundedRegExp < >, automaton::ExtendedNFA < > >(ToRegExpStateElimination::convert);
auto ToRegExpStateEliminationCompactNFA = ToRegExpStateElimination::RegistratorWrapper<regexp::UnboundedRegExp < >, automaton::CompactNFA < > >(ToRegExpStateElimination::convert);

automaton::ExtendedNFA < > ToRegExpStateElimination::eliminateState(const automaton::ExtendedNFA < > & extendedAutomaton, const label::Label& q) {
	automaton::ExtendedNFA < > newAutomaton(extendedAutomaton.getInitialState()); // sure that q is neither initial nor final (follows from step 2 - extending ExtendedNFA)
	newAutomaton.setStates(extendedAutomaton.getStates());
	newAutomaton.removeState(q); // preserve all states but q (the one to eliminate)
	newAutomaton.setInputAlphabet(extendedAutomaton.getInputAlphabet());
	newAutomaton.setFinalStates(extendedAutomaton.getFinalStates());

	for(const auto& p: newAutomaton.getStates()) {
		for(const auto& r : newAutomaton.getStates()) {
			regexp::UnboundedRegExpStructure < alphabet::Symbol > concat = transition(extendedAutomaton, p, q);
			concat = regexp::RegExpConcatenate::concatenate(concat, regexp::RegExpIterate::iterate(transition(extendedAutomaton, q, q)));
			concat = regexp::RegExpConcatenate::concatenate(concat, transition(extendedAutomaton, q, r));

			regexp::UnboundedRegExpStructure < alphabet::Symbol > alt = regexp::RegExpAlternate::alternate(concat, transition(extendedAutomaton, p, r));

			newAutomaton.addTransition(p, regexp::simplify::RegExpOptimize::optimize(alt), r);
		}
	}

	return newAutomaton;
}

const regexp::UnboundedRegExpStructure < alphabet::Symbol > ToRegExpStateElimination::transition(const automaton::ExtendedNFA < > & automaton, const label::Label& from, const label::Label& to) {
	regexp::UnboundedRegExpStructure < alphabet::Symbol > ret(regexp::UnboundedRegExpEmpty < alphabet::Symbol > { });

	for(const auto& transition: automaton.getTransitionsFromState(from))
		if(transition.second.count(to) > 0)
			ret = regexp::RegExpAlternate::alternate(ret, transition.first.second);

	return regexp::simplify::RegExpOptimize::optimize(ret);
}

void ToRegExpStateElimination::extendExtendedNFA(automaton::ExtendedNFA < > & automaton) {
	const label::Label& initState = automaton.getInitialState();
	if(automaton.getFinalStates().count(initState) > 0 || automaton.getTransitionsToState(initState).size() > 0 ) {
		label::Label q0 = common::createUnique(initState, automaton.getStates());
		automaton.addState(q0);

		regexp::UnboundedRegExpStructure < alphabet::Symbol > regexp { regexp::UnboundedRegExpEpsilon < alphabet::Symbol > () };
		automaton.addTransition(q0, regexp, initState);

		automaton.setInitialState(q0);
	}

	if(automaton.getFinalStates().size() > 1) {
		label::Label f = common::createUnique(label::FinalStateLabel::instance < label::Label > ( ), automaton.getStates());
		automaton.addState(f);

		for(const auto &state : automaton.getFinalStates() ) {
			regexp::UnboundedRegExpStructure < alphabet::Symbol > regexp { regexp::UnboundedRegExpEpsilon < alphabet::Symbol > () };
			automaton.addTransition(state, regexp, f);

			automaton.removeFinalState(state);
		}

		automaton.addFinalState(f);
	}
}

} /* namespace convert */

} /* namespace automaton */
